# © 2019-present nextmv.io inc

"""
Methods for solving a Vehicle Routing Problem with Nextroute.
"""

import json
import os
import platform
import subprocess
from typing import Any, Dict, Union

from nextroute.options import Options
from nextroute.schema.input import Input
from nextroute.schema.output import Output

SUPPORTED_OS = ["linux", "windows", "darwin"]
"""The operating systems supported by the Nextroute engine."""
SUPPORTED_ARCHITECTURES = ["amd64", "x86_64", "arm64", "aarch64"]
"""The architectures supported by the Nextroute engine."""


def solve(
    input: Union[Input, Dict[str, Any]],
    options: Union[Options, Dict[str, Any]],
) -> Output:
    """
    Solve a Vehicle Routing Problem (VRP) using the Nextroute engine. The input
    and options are passed to the engine, and the output is returned. The input
    and options can be provided as dictionaries or as objects, although the
    recommended way is to use the classes, as they provide validation.

    Examples
    --------

    * Using default options to load an input from a file.
        ```python
        import json

        import nextroute

        with open("input.json") as f:
            data = json.load(f)

        input = nextroute.schema.Input.from_dict(data)
        options = nextroute.Options()
        output = nextroute.solve(input, options)
        print(output)
        ```

    * Using custom options to load an input from a file.
        ```python
        import json

        import nextroute

        with open("input.json") as f:
            data = json.load(f)

        input = nextroute.schema.Input.from_dict(data)
        options = nextroute.Options(
            solve=nextroute.ParallelSolveOptions(duration=2),
        )
        output = nextroute.solve(input, options)
        print(output)
        ```

    * Using custom dict options to load an input from a file.
        ```python
        import json

        import nextroute

        with open("input.json") as f:
            data = json.load(f)

        input = nextroute.schema.Input.from_dict(data)
        options = {
            "solve": {
                "duration": 2,
            },
        }
        output = nextroute.solve(input, options)
        print(output)
        ```


    Parameters
    ----------
    input : Union[schema.Input, Dict[str, Any]]
        The input to the Nextroute engine. If a dictionary is provided, it will
        be converted to an Input object to validate it.
    options : Union[Options, Dict[str, Any]]
        The options for the Nextroute engine. If a dictionary is provided, it
        will be converted to an Options object.

    Returns
    -------
    schema.Output
        The output of the Nextroute engine. You can call the `to_dict` method
        on this object to get a dictionary representation of the output.
    """

    if isinstance(input, dict):
        input = Input.from_dict(input)

    input_stream = json.dumps(input.to_dict())

    if isinstance(options, dict):
        options = Options.from_dict(options)

    os_name = platform.system().lower()
    if os_name not in SUPPORTED_OS:
        raise Exception(f'unsupported operating system: "{os_name}", supported os are: {", ".join(SUPPORTED_OS)}')

    architecture = platform.machine().lower()
    if architecture not in SUPPORTED_ARCHITECTURES:
        raise Exception(
            f'unsupported architecture: "{architecture}", supported arch are: {", ".join(SUPPORTED_ARCHITECTURES)}'
        )

    executable = os.path.join(os.path.dirname(__file__), "bin", "nextroute.exe")
    if not os.path.exists(executable):
        raise Exception(f"missing Nextroute binary: {executable}")

    option_args = options.to_args()
    args = [executable] + option_args

    try:
        result = subprocess.run(
            args,
            env=os.environ,
            check=True,
            text=True,
            capture_output=True,
            input=input_stream,
        )

    except subprocess.CalledProcessError as e:
        raise Exception(f"error running Nextroute binary: {e.stderr}") from e

    raw_output = result.stdout
    output = Output.from_dict(json.loads(raw_output))

    return output



# © 2019-present nextmv.io inc

"""
Defines the location class.
"""

from nextroute.base_model import BaseModel


class Location(BaseModel):
    """Location represents a geographical location."""

    lat: float
    """Latitude of the location."""
    lon: float
    """Longitude of the location."""



# © 2019-present nextmv.io inc

import json
import math
import os
import unittest

from nextroute import check as nextrouteCheck
from nextroute.schema import (
    Location,
    ObjectiveOutput,
    Output,
    PlannedStopOutput,
    ResultStatistics,
    RunStatistics,
    SeriesData,
    Solution,
    Statistics,
    StopOutput,
    VehicleOutput,
    Version,
)


class TestOutput(unittest.TestCase):
    filepath = os.path.join(os.path.dirname(__file__), "output.json")

    def test_from_json(self):
        with open(self.filepath) as f:
            json_data = json.load(f)

        nextroute_output = Output.from_dict(json_data)
        parsed = nextroute_output.to_dict()
        solution = parsed["solutions"][0]

        for s, stop in enumerate(solution["unplanned"]):
            original_stop = json_data["solutions"][0]["unplanned"][s]
            self.assertEqual(
                stop,
                original_stop,
                f"stop: parsed({stop}) and original ({original_stop}) should be equal",
            )

        for v, vehicle in enumerate(solution["vehicles"]):
            original_vehicle = json_data["solutions"][0]["vehicles"][v]
            self.assertEqual(
                vehicle,
                original_vehicle,
                f"vehicle: parsed ({vehicle}) and original ({original_vehicle}) should be equal",
            )

        self.assertEqual(
            solution["objective"],
            json_data["solutions"][0]["objective"],
            f"objective: parsed ({solution['objective']}) and "
            f"original ({json_data['solutions'][0]['objective']}) should be equal",
        )

        statistics = parsed["statistics"]
        self.assertEqual(
            statistics["run"],
            json_data["statistics"]["run"],
            f"run: parsed ({statistics['run']}) and original ({json_data['statistics']['run']}) should be equal",
        )
        self.assertEqual(
            statistics["result"],
            json_data["statistics"]["result"],
            f"result: parsed ({statistics['result']}) and "
            f"original ({json_data['statistics']['result']}) should be equal",
        )

    def test_from_dict(self):
        with open(self.filepath) as f:
            json_data = json.load(f)

        nextroute_output = Output.from_dict(json_data)

        version = nextroute_output.version
        self.assertTrue(isinstance(version, Version), "Version should be of type Version.")

        solutions = nextroute_output.solutions
        for solution in solutions:
            self.assertTrue(
                isinstance(solution, Solution),
                f"Solution ({solution}) should be of type Solution.",
            )

            unplanned = solution.unplanned
            for stop in unplanned:
                self.assertTrue(
                    isinstance(stop, StopOutput),
                    f"Stop ({stop}) should be of type StopOutput.",
                )
                self.assertTrue(
                    stop.id is not None,
                    f"Stop ({stop}) should have an id.",
                )
                self.assertNotEqual(
                    stop.id,
                    "",
                    f"Stop ({stop}) should have a valid id.",
                )
                self.assertTrue(
                    isinstance(stop.location, Location),
                    f"Stop ({stop}) should have a location.",
                )
                self.assertGreaterEqual(
                    stop.location.lat,
                    -90,
                    f"Stop ({stop}) should have a valid latitude.",
                )
                self.assertLessEqual(
                    stop.location.lat,
                    90,
                    f"Stop ({stop}) should have a valid latitude.",
                )
                self.assertGreaterEqual(
                    stop.location.lon,
                    -180,
                    f"Stop ({stop}) should have a valid longitude.",
                )
                self.assertLessEqual(
                    stop.location.lon,
                    180,
                    f"Stop ({stop}) should have a valid longitude.",
                )

            vehicles = solution.vehicles
            for vehicle in vehicles:
                self.assertTrue(
                    isinstance(vehicle, VehicleOutput),
                    f"Vehicle ({vehicle}) should be of type VehicleOutput.",
                )
                self.assertTrue(
                    vehicle.id is not None,
                    f"Vehicle ({vehicle}) should have an id.",
                )
                self.assertNotEqual(
                    vehicle.id,
                    "",
                    f"Vehicle ({vehicle}) should have a valid id.",
                )
                self.assertGreaterEqual(
                    vehicle.route_duration,
                    0,
                    f"Vehicle ({vehicle}) should have a valid route duration.",
                )
                self.assertGreaterEqual(
                    vehicle.route_stops_duration,
                    0,
                    f"Vehicle ({vehicle}) should have a valid route stops duration.",
                )
                self.assertGreaterEqual(
                    vehicle.route_travel_distance,
                    0,
                    f"Vehicle ({vehicle}) should have a valid route travel distance.",
                )
                self.assertGreaterEqual(
                    vehicle.route_travel_duration,
                    0,
                    f"Vehicle ({vehicle}) should have a valid route travel duration.",
                )

                for stop in vehicle.route:
                    self.assertTrue(
                        isinstance(stop, PlannedStopOutput),
                        f"Stop ({stop}) should be of type PlannedStopOutput.",
                    )
                    self.assertTrue(
                        isinstance(stop.stop, StopOutput),
                        f"Stop ({stop}) should have a stop.",
                    )
                    self.assertGreaterEqual(
                        stop.travel_duration,
                        0,
                        f"Stop ({stop}) should have a valid travel duration.",
                    )
                    self.assertGreaterEqual(
                        stop.cumulative_travel_duration,
                        0,
                        f"Stop ({stop}) should have a valid cumulative travel duration.",
                    )

            objective = solution.objective
            self.assertTrue(
                isinstance(objective, ObjectiveOutput),
                f"Objective ({objective}) should be of type ObjectiveOutput.",
            )

        statistics = nextroute_output.statistics
        self.assertTrue(
            isinstance(statistics, Statistics),
            f"Statistics ({statistics}) should be of type Statistics.",
        )

        run_statistics = statistics.run
        self.assertTrue(
            isinstance(run_statistics, RunStatistics),
            f"Run statistics ({run_statistics}) should be of type RunStatistics.",
        )
        self.assertGreaterEqual(
            run_statistics.duration,
            0,
            f"Run statistics ({run_statistics}) should have a valid duration.",
        )
        self.assertGreaterEqual(
            run_statistics.iterations,
            0,
            f"Run statistics ({run_statistics}) should have a valid number of iterations.",
        )

        result_statistics = statistics.result
        self.assertTrue(
            isinstance(result_statistics, ResultStatistics),
            f"Result statistics ({result_statistics}) should be of type ResultStatistics.",
        )
        self.assertGreaterEqual(
            result_statistics.duration,
            0,
            f"Result statistics ({result_statistics}) should have a valid duration.",
        )
        self.assertGreaterEqual(
            result_statistics.value,
            0,
            f"Result statistics ({result_statistics}) should have a valid value.",
        )

        series_data = statistics.series_data
        self.assertTrue(
            isinstance(series_data, SeriesData),
            f"Series data ({series_data}) should be of type SeriesData.",
        )

    def test_with_check(self):
        with open(os.path.join(os.path.dirname(__file__), "output_with_check.json")) as f:
            json_data = json.load(f)

        nextroute_output = Output.from_dict(json_data)
        check = nextroute_output.solutions[0].check
        self.assertTrue(
            isinstance(check, nextrouteCheck.Output),
            f"Check ({check}) should be of type nextrouteCheck.Output.",
        )
        self.assertTrue(
            isinstance(check.solution, nextrouteCheck.Solution),
            f"Solution ({check.solution}) should be of type nextrouteCheck.checkSolution.",
        )
        self.assertTrue(
            isinstance(check.summary, nextrouteCheck.Summary),
            f"Summary ({check.summary}) should be of type nextrouteCheck.Summary.",
        )
        for plan_unit in check.plan_units:
            self.assertTrue(
                isinstance(plan_unit, nextrouteCheck.PlanUnit),
                f"Plan unit ({plan_unit}) should be of type nextrouteCheck.PlanUnit.",
            )

        for vehicle in check.vehicles:
            self.assertTrue(
                isinstance(vehicle, nextrouteCheck.Vehicle),
                f"Vehicle ({vehicle}) should be of type nextrouteCheck.Vehicle",
            )

    def test_result_statistics_decoding(self):
        test_cases = [
            {
                "name": "value is float",
                "json_stats": '{"duration": 0.1, "value": 1.23}',
            },
            {
                "name": "value is nan",
                "json_stats": '{"duration": 0.1, "value": "nan"}',
            },
            {
                "name": "value is infinity",
                "json_stats": '{"duration": 0.1, "value": "inf"}',
            },
            {
                "name": "value is infinity 2",
                "json_stats": '{"duration": 0.1, "value": "+inf"}',
            },
            {
                "name": "value is -infinity",
                "json_stats": '{"duration": 0.1, "value": "-inf"}',
            },
        ]

        for test in test_cases:
            dict_stats = json.loads(test["json_stats"])
            stats = ResultStatistics.from_dict(dict_stats)
            self.assertTrue(isinstance(stats, ResultStatistics))
            self.assertTrue(isinstance(stats.value, float))

    def test_result_statistics_encoding(self):
        test_cases = [
            {
                "name": "value is float",
                "stats": ResultStatistics(duration=0.1, value=1.23),
            },
            {
                "name": "value is nan",
                "stats": ResultStatistics(duration=0.1, value=math.nan),
            },
            {
                "name": "value is infinity",
                "stats": ResultStatistics(duration=0.1, value=math.inf),
            },
            {
                "name": "value is -infinity",
                "stats": ResultStatistics(duration=0.1, value=-1 * math.inf),
            },
        ]

        for test in test_cases:
            stats = test["stats"]
            dict_stats = stats.to_dict()
            self.assertTrue(isinstance(dict_stats, dict))
            self.assertTrue(isinstance(dict_stats["value"], float))



# © 2019-present nextmv.io inc

"""
Options for working with the Nextroute engine.
"""

import json
from enum import Enum
from typing import Any, Dict, List

from pydantic import Field

from nextroute.base_model import BaseModel

# Arguments that require a duration suffix.
_DURATIONS_ARGS = [
    "-check.duration",
    "-solve.duration",
]

# Arguments that require a string enum.
_STR_ENUM_ARGS = [
    "CHECK_VERBOSITY",
]


class Verbosity(str, Enum):
    """Format of an `Input`."""

    OFF = "off"
    """The check engine is not run."""
    LOW = "low"
    """Low verbosity for the check engine."""
    MEDIUM = "medium"
    """Medium verbosity for the check engine."""
    HIGH = "high"
    """High verbosity for the check engine."""


class Options(BaseModel):
    """Options for using Nextroute."""

    CHECK_DURATION: float = 30
    """Maximum duration of the check, in seconds."""
    CHECK_VERBOSITY: Verbosity = Verbosity.OFF
    """Verbosity of the check engine."""
    FORMAT_DISABLE_PROGRESSION: bool = False
    """Whether to disable the progression series."""
    MODEL_CONSTRAINTS_DISABLE_ATTRIBUTES: bool = False
    """Ignore the compatibility attributes constraint."""
    MODEL_CONSTRAINTS_DISABLE_CAPACITIES: List[str] = Field(default_factory=list)
    """Ignore the capacity constraint for the given resource names."""
    MODEL_CONSTRAINTS_DISABLE_CAPACITY: bool = False
    """Ignore the capacity constraint for all resources."""
    MODEL_CONSTRAINTS_DISABLE_DISTANCELIMIT: bool = False
    """Ignore the distance limit constraint."""
    MODEL_CONSTRAINTS_DISABLE_GROUPS: bool = False
    """Ignore the groups constraint."""
    MODEL_CONSTRAINTS_DISABLE_MAXIMUMDURATION: bool = False
    """Ignore the maximum duration constraint."""
    MODEL_CONSTRAINTS_DISABLE_MAXIMUMSTOPS: bool = False
    """Ignore the maximum stops constraint."""
    MODEL_CONSTRAINTS_DISABLE_MAXIMUMWAITSTOP: bool = False
    """Ignore the maximum stop wait constraint."""
    MODEL_CONSTRAINTS_DISABLE_MAXIMUMWAITVEHICLE: bool = False
    """Ignore the maximum vehicle wait constraint."""
    MODEL_CONSTRAINTS_DISABLE_MIXINGITEMS: bool = False
    """Ignore the do not mix items constraint."""
    MODEL_CONSTRAINTS_DISABLE_PRECEDENCE: bool = False
    """Ignore the precedence (pickups & deliveries) constraint."""
    MODEL_CONSTRAINTS_DISABLE_STARTTIMEWINDOWS: bool = False
    """Ignore the start time windows constraint."""
    MODEL_CONSTRAINTS_DISABLE_VEHICLEENDTIME: bool = False
    """Ignore the vehicle end time constraint."""
    MODEL_CONSTRAINTS_DISABLE_VEHICLESTARTTIME: bool = False
    """Ignore the vehicle start time constraint."""
    MODEL_CONSTRAINTS_ENABLE_CLUSTER: bool = False
    """Enable the cluster constraint."""
    MODEL_OBJECTIVES_CAPACITIES: str = ""
    """
    Capacity objective, provide triple for each resource
    `name:default;factor:1.0;offset;0.0`.
    """
    MODEL_OBJECTIVES_CLUSTER: float = 0.0
    """Factor to weigh the cluster objective."""
    MODEL_OBJECTIVES_EARLYARRIVALPENALTY: float = 1.0
    """Factor to weigh the early arrival objective."""
    MODEL_OBJECTIVES_LATEARRIVALPENALTY: float = 1.0
    """Factor to weigh the late arrival objective."""
    MODEL_OBJECTIVES_MINSTOPS: float = 1.0
    """Factor to weigh the min stops objective."""
    MODEL_OBJECTIVES_TRAVELDURATION: float = 0.0
    """Factor to weigh the travel duration objective."""
    MODEL_OBJECTIVES_UNPLANNEDPENALTY: float = 1.0
    """Factor to weigh the unplanned objective."""
    MODEL_OBJECTIVES_VEHICLEACTIVATIONPENALTY: float = 1.0
    """Factor to weigh the vehicle activation objective."""
    MODEL_OBJECTIVES_VEHICLESDURATION: float = 1.0
    """Factor to weigh the vehicles duration objective."""
    MODEL_PROPERTIES_DISABLE_DURATIONGROUPS: bool = False
    """Ignore the durations groups of stops."""
    MODEL_PROPERTIES_DISABLE_DURATIONS: bool = False
    """Ignore the durations of stops."""
    MODEL_PROPERTIES_DISABLE_INITIALSOLUTION: bool = False
    """Ignore the initial solution."""
    MODEL_PROPERTIES_DISABLE_STOPDURATIONMULTIPLIERS: bool = False
    """Ignore the stop duration multipliers defined on vehicles."""
    MODEL_VALIDATE_DISABLE_RESOURCES: bool = False
    """Disable the resources validation."""
    MODEL_VALIDATE_DISABLE_STARTTIME: bool = False
    """Disable the start time validation."""
    MODEL_VALIDATE_ENABLE_MATRIX: bool = False
    """Enable matrix validation."""
    MODEL_VALIDATE_ENABLE_MATRIXASYMMETRYTOLERANCE: int = 20
    """Percentage of acceptable matrix asymmetry, requires matrix validation enabled."""
    SOLVE_DURATION: float = 5
    """Maximum duration, in seconds, of the solver."""
    SOLVE_ITERATIONS: int = -1
    """
    Maximum number of iterations, -1 assumes no limit; iterations are counted
    after start solutions are generated.
    """
    SOLVE_PARALLELRUNS: int = -1
    """
    Maximum number of parallel runs, -1 results in using all available
    resources.
    """
    SOLVE_RUNDETERMINISTICALLY: bool = False
    """Run the parallel solver deterministically."""
    SOLVE_STARTSOLUTIONS: int = -1
    """
    Number of solutions to generate on top of those passed in; one solution
    generated with sweep algorithm, the rest generated randomly.
    """

    def to_args(self) -> List[str]:
        """
        Convert the options to command-line arguments.

        Returns
        ----------
        List[str]
            The flattened options as a list of strings.
        """

        opt_dict = self.to_dict()

        default_options = Options()
        default_options_dict = default_options.to_dict()

        args = []
        for key, value in opt_dict.items():
            # We only care about custom options, so we skip the default ones.
            default_value = default_options_dict.get(key)
            if value == default_value:
                continue

            key = f"-{key.replace('_', '.').lower()}"

            str_value = json.dumps(value)
            if key in _DURATIONS_ARGS:
                str_value = str_value + "s"  # Transforms into seconds.

            if str_value.startswith('"') and str_value.endswith('"'):
                str_value = str_value[1:-1]

            # Nextroute’s Go implementation does not support boolean flags with
            # values. If the value is a boolean, then we only append the key if
            # the value is True.
            should_append_value = True
            if isinstance(value, bool):
                if not value:
                    continue

                should_append_value = False

            args.append(key)
            if should_append_value:
                args.append(str_value)

        return args

    @classmethod
    def extract_from_dict(cls, data: Dict[str, Any]) -> "Options":
        """
        Extracts options from a dictionary. This dictionary may contain more
        keys that are not part of the Nextroute options.

        Parameters
        ----------
        data : Dict[str, Any]
            The dictionary to extract options from.

        Returns
        ----------
        Options
            The Nextroute options.
        """

        options = cls()
        for key, value in data.items():
            key = key.upper()
            if not hasattr(options, key):
                continue

            # Enums need to be handled manually.
            if key == "CHECK_VERBOSITY":
                value = Verbosity(value)

            setattr(options, key, value)

        return options



# © 2019-present nextmv.io inc

import os
import subprocess


def nextroute_version() -> str:
    """
    Get the version of the embedded Nextroute binary.
    """
    executable = os.path.join(os.path.dirname(__file__), "bin", "nextroute.exe")
    return subprocess.check_output([executable, "--version"]).decode().strip()



# © 2019-present nextmv.io inc

"""
Schema (class) definitions for the entities in Nextroute.
"""

from .input import Defaults as Defaults
from .input import DurationGroup as DurationGroup
from .input import Input as Input
from .location import Location as Location
from .output import ObjectiveOutput as ObjectiveOutput
from .output import Output as Output
from .output import PlannedStopOutput as PlannedStopOutput
from .output import Solution as Solution
from .output import StopOutput as StopOutput
from .output import VehicleOutput as VehicleOutput
from .output import Version as Version
from .statistics import DataPoint as DataPoint
from .statistics import ResultStatistics as ResultStatistics
from .statistics import RunStatistics as RunStatistics
from .statistics import Series as Series
from .statistics import SeriesData as SeriesData
from .statistics import Statistics as Statistics
from .stop import AlternateStop as AlternateStop
from .stop import Stop as Stop
from .stop import StopDefaults as StopDefaults
from .vehicle import InitialStop as InitialStop
from .vehicle import Vehicle as Vehicle
from .vehicle import VehicleDefaults as VehicleDefaults



# © 2019-present nextmv.io inc

"""
Defines the stop class.
"""

from datetime import datetime
from typing import Any, List, Optional

from nextroute.base_model import BaseModel
from nextroute.schema.location import Location


class StopDefaults(BaseModel):
    """Default values for a stop."""

    compatibility_attributes: Optional[List[str]] = None
    """Attributes that the stop is compatible with."""
    duration: Optional[int] = None
    """Duration of the stop in seconds."""
    early_arrival_time_penalty: Optional[float] = None
    """Penalty per second for arriving at the stop before the target arrival time."""
    late_arrival_time_penalty: Optional[float] = None
    """Penalty per second for arriving at the stop after the target arrival time."""
    max_wait: Optional[int] = None
    """Maximum waiting duration in seconds at the stop."""
    quantity: Optional[Any] = None
    """Quantity of the stop."""
    start_time_window: Optional[Any] = None
    """Time window in which the stop can start service."""
    target_arrival_time: Optional[datetime] = None
    """Target arrival time at the stop."""
    unplanned_penalty: Optional[int] = None
    """Penalty for not planning a stop."""


class Stop(StopDefaults):
    """Stop is a location that must be visited by a vehicle in a Vehicle
    Routing Problem (VRP.)"""

    id: str
    """Unique identifier for the stop."""
    location: Location
    """Location of the stop."""

    custom_data: Optional[Any] = None
    """Arbitrary data associated with the stop."""
    mixing_items: Optional[Any] = None
    """Defines the items that are inserted or removed from the vehicle when visiting the stop."""
    precedes: Optional[Any] = None
    """Stops that must be visited after this one on the same route."""
    succeeds: Optional[Any] = None
    """Stops that must be visited before this one on the same route."""


class AlternateStop(StopDefaults):
    """An alternate stop can be serviced instead of another stop."""

    id: str
    """Unique identifier for the stop."""
    location: Location
    """Location of the stop."""

    custom_data: Optional[Any] = None
    """Arbitrary data associated with the stop."""



# © 2019-present nextmv.io inc

"""
JSON class for data wrangling JSON objects.
"""

from typing import Any, Dict

from pydantic import BaseModel


class BaseModel(BaseModel):
    """Base class for data wrangling tasks with JSON."""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Instantiates the class from a dict."""

        return cls(**data)

    def to_dict(self) -> Dict[str, Any]:
        """Converts the class to a dict."""

        return self.model_dump(mode="json", exclude_none=True, by_alias=True)



# This script is copied to the `src` root so that the `nextroute` import is
# resolved. It is fed an input via stdin and is meant to write the output to
# stdout.
from typing import Any, Dict

import nextmv

import nextroute


def main() -> None:
    """Entry point for the program."""

    parameters = [
        nextmv.Parameter("input", str, "", "Path to input file. Default is stdin.", False),
        nextmv.Parameter("output", str, "", "Path to output file. Default is stdout.", False),
    ]

    nextroute_options = nextroute.Options()
    for name, default_value in nextroute_options.to_dict().items():
        parameters.append(nextmv.Parameter(name.lower(), type(default_value), default_value, name, False))

    options = nextmv.Options(*parameters)

    input = nextmv.load_local(options=options, path=options.input)

    nextmv.log("Solving vehicle routing problem:")
    nextmv.log(f"  - stops: {len(input.data.get('stops', []))}")
    nextmv.log(f"  - vehicles: {len(input.data.get('vehicles', []))}")

    output = solve(input, options)
    nextmv.write_local(output, path=options.output)


def solve(input: nextmv.Input, options: nextmv.Options) -> Dict[str, Any]:
    """Solves the given problem and returns the solution."""

    nextroute_input = nextroute.schema.Input.from_dict(input.data)
    nextroute_options = nextroute.Options.extract_from_dict(options.to_dict())
    nextroute_output = nextroute.solve(nextroute_input, nextroute_options)

    return nextroute_output.to_dict()


if __name__ == "__main__":
    main()



# © 2019-present nextmv.io inc

"""
The Nextroute Python interface.

Nextroute is a flexible engine for solving Vehicle Routing Problems (VRPs). The
core of Nextroute is written in Go and this package provides a Python interface
to it.
"""

from .__about__ import __version__
from .options import Options as Options
from .options import Verbosity as Verbosity
from .solve import solve as solve
from .version import nextroute_version as nextroute_version

VERSION = __version__
"""The version of the Nextroute Python package."""



# © 2019-present nextmv.io inc

"""
Defines the output class.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional

from nextroute.base_model import BaseModel
from nextroute.check.schema import Output as CheckOutput
from nextroute.schema.location import Location
from nextroute.schema.statistics import Statistics


class Version(BaseModel):
    """A version used for solving."""

    sdk: str
    """Nextmv SDK."""


class StopOutput(BaseModel):
    """Basic structure for the output of a stop."""

    id: str
    """ID of the stop."""
    location: Location
    """Location of the stop."""

    custom_data: Optional[Any] = None
    """Custom data of the stop."""


class PlannedStopOutput(BaseModel):
    """Output of a stop planned in the solution."""

    stop: StopOutput
    """Basic information on the stop."""

    arrival_time: Optional[datetime] = None
    """Actual arrival time at this stop."""
    cumulative_travel_distance: Optional[float] = None
    """Cumulative distance to travel from the first stop to this one, in meters."""
    cumulative_travel_duration: Optional[float] = None
    """Cumulative duration to travel from the first stop to this one, in seconds."""
    custom_data: Optional[Any] = None
    """Custom data of the stop."""
    duration: Optional[float] = None
    """Duration of the service at the stop, in seconds."""
    early_arrival_duration: Optional[float] = None
    """Duration of early arrival at the stop, in seconds."""
    end_time: Optional[datetime] = None
    """End time of the service at the stop."""
    late_arrival_duration: Optional[float] = None
    """Duration of late arrival at the stop, in seconds."""
    mix_items: Optional[Any] = None
    """Mix items at the stop."""
    start_time: Optional[datetime] = None
    """Start time of the service at the stop."""
    target_arrival_time: Optional[datetime] = None
    """Target arrival time at this stop."""
    travel_distance: Optional[float] = None
    """Distance to travel from the previous stop to this one, in meters."""
    travel_duration: Optional[float] = None
    """Duration to travel from the previous stop to this one, in seconds."""
    waiting_duration: Optional[float] = None
    """Waiting duratino at the stop, in seconds."""


class VehicleOutput(BaseModel):
    """Output of a vehicle in the solution."""

    id: str
    """ID of the vehicle."""

    alternate_stops: Optional[List[str]] = None
    """List of alternate stops that were planned on the vehicle."""
    custom_data: Optional[Any] = None
    """Custom data of the vehicle."""
    route: Optional[List[PlannedStopOutput]] = None
    """Route of the vehicle, which is a list of stops that were planned on
    it."""
    route_duration: Optional[float] = None
    """Total duration of the vehicle's route, in seconds."""
    route_stops_duration: Optional[float] = None
    """Total duration of the stops of the vehicle, in seconds."""
    route_travel_distance: Optional[float] = None
    """Total travel distance of the vehicle, in meters."""
    route_travel_duration: Optional[float] = None
    """Total travel duration of the vehicle, in seconds."""
    route_waiting_duration: Optional[float] = None
    """Total waiting duration of the vehicle, in seconds."""


class ObjectiveOutput(BaseModel):
    """Information of the objective (value function)."""

    name: str
    """Name of the objective."""

    base: Optional[float] = None
    """Base of the objective."""
    custom_data: Optional[Any] = None
    """Custom data of the objective."""
    factor: Optional[float] = None
    """Factor of the objective."""
    objectives: Optional[List[Dict[str, Any]]] = None
    """List of objectives. Each list is actually of the same class
    `ObjectiveOutput`, but we avoid a recursive definition here."""
    value: Optional[float] = None
    """Value of the objective, which is equivalent to `self.base *
    self.factor`."""


class Solution(BaseModel):
    """Solution to a Vehicle Routing Problem (VRP)."""

    unplanned: Optional[List[StopOutput]] = None
    """List of stops that were not planned in the solution."""
    vehicles: Optional[List[VehicleOutput]] = None
    """List of vehicles in the solution."""
    objective: Optional[ObjectiveOutput] = None
    """Information of the objective (value function)."""
    check: Optional[CheckOutput] = None
    """Check of the solution, if enabled."""


class Output(BaseModel):
    """Output schema for Nextroute."""

    options: Dict[str, Any]
    """Options used to obtain this output."""
    version: Version
    """Versions used for the solution."""

    solutions: Optional[List[Solution]] = None
    """Solutions to the problem."""
    statistics: Optional[Statistics] = None
    """Statistics of the solution."""



# © 2019-present nextmv.io inc

"""
Defines the input class.
"""

from datetime import datetime
from typing import Any, List, Optional, Union

from nextroute.base_model import BaseModel
from nextroute.schema.stop import AlternateStop, Stop, StopDefaults
from nextroute.schema.vehicle import Vehicle, VehicleDefaults


class Defaults(BaseModel):
    """Default values for vehicles and stops."""

    stops: Optional[StopDefaults] = None
    """Default values for stops."""
    vehicles: Optional[VehicleDefaults] = None
    """Default values for vehicles."""


class DurationGroup(BaseModel):
    """Represents a group of stops that get additional duration whenever a stop
    of the group is approached for the first time."""

    duration: int
    """Duration to add when visiting the group."""
    group: List[str]
    """Stop IDs contained in the group."""


class MatrixTimeFrame(BaseModel):
    """Represents a time-dependent duration matrix or scaling factor."""

    start_time: datetime
    """Start time of the time frame."""
    end_time: datetime
    """End time of the time frame."""
    matrix: Optional[List[List[float]]] = None
    """Duration matrix for the time frame."""
    scaling_factor: Optional[float] = None
    """Scaling factor for the time frame."""


class TimeDependentMatrix(BaseModel):
    """Represents time-dependent duration matrices."""

    vehicle_ids: Optional[List[str]] = None
    """Vehicle IDs for which the duration matrix is defined."""
    default_matrix: List[List[float]]
    """Default duration matrix."""
    matrix_time_frames: Optional[List[MatrixTimeFrame]] = None
    """Time-dependent duration matrices."""


class Input(BaseModel):
    """Input schema for Nextroute."""

    stops: List[Stop]
    """Stops that must be visited by the vehicles."""
    vehicles: List[Vehicle]
    """Vehicles that service the stops."""

    alternate_stops: Optional[List[AlternateStop]] = None
    """A set of alternate stops for the vehicles."""
    custom_data: Optional[Any] = None
    """Arbitrary data associated with the input."""
    defaults: Optional[Defaults] = None
    """Default values for vehicles and stops."""
    distance_matrix: Optional[List[List[float]]] = None
    """Matrix of travel distances in meters between stops."""
    duration_groups: Optional[List[DurationGroup]] = None
    """Duration in seconds added when approaching the group."""
    duration_matrix: Optional[
        Union[
            List[List[float]],
            TimeDependentMatrix,
            List[TimeDependentMatrix],
        ]
    ] = None
    """Matrix of travel durations in seconds between stops as a single matrix or duration matrices."""
    options: Optional[Any] = None
    """Arbitrary options."""
    stop_groups: Optional[List[List[str]]] = None
    """Groups of stops that must be part of the same route."""



# © 2019-present nextmv.io inc

"""
Defines the vehicle class.
"""

from datetime import datetime
from typing import Any, List, Optional

from nextroute.base_model import BaseModel
from nextroute.schema.location import Location


class InitialStop(BaseModel):
    """Represents a stop that is already planned on a vehicle."""

    id: str
    """Unique identifier of the stop."""

    fixed: Optional[bool] = None
    """Whether the stop is fixed on the route."""


class VehicleDefaults(BaseModel):
    """Default values for vehicles."""

    activation_penalty: Optional[int] = None
    """Penalty of using the vehicle."""
    alternate_stops: Optional[List[str]] = None
    """A set of alternate stops for which only one should be serviced."""
    capacity: Optional[Any] = None
    """Capacity of the vehicle."""
    compatibility_attributes: Optional[List[str]] = None
    """Attributes that the vehicle is compatible with."""
    end_location: Optional[Location] = None
    """Location where the vehicle ends."""
    end_time: Optional[datetime] = None
    """Latest time at which the vehicle ends its route."""
    max_distance: Optional[int] = None
    """Maximum distance in meters that the vehicle can travel."""
    max_duration: Optional[int] = None
    """Maximum duration in seconds that the vehicle can travel."""
    max_stops: Optional[int] = None
    """Maximum number of stops that the vehicle can visit."""
    max_wait: Optional[int] = None
    """Maximum aggregated waiting time that the vehicle can wait across route stops."""
    min_stops: Optional[int] = None
    """Minimum stops that a vehicle should visit."""
    min_stops_penalty: Optional[float] = None
    """Penalty for not visiting the minimum number of stops."""
    speed: Optional[float] = None
    """Speed of the vehicle in meters per second."""
    start_level: Optional[Any] = None
    """Initial level of the vehicle."""
    start_location: Optional[Location] = None
    """Location where the vehicle starts."""
    start_time: Optional[datetime] = None
    """Time when the vehicle starts its route."""


class Vehicle(VehicleDefaults):
    """A vehicle services stops in a Vehicle Routing Problem (VRP)."""

    id: str
    """Unique identifier of the vehicle."""

    custom_data: Optional[Any] = None
    """Arbitrary custom data."""
    initial_stops: Optional[List[InitialStop]] = None
    """Initial stops planned on the vehicle."""
    stop_duration_multiplier: Optional[float] = None
    """Multiplier for the duration of stops."""



# © 2019-present nextmv.io inc

"""
Schema for statistics.
"""

from typing import Any, Dict, List, Optional, Union

from pydantic import Field

from nextroute.base_model import BaseModel


class RunStatistics(BaseModel):
    """
    Statistics about a general run.

    Parameters
    ----------
    duration : float, optional
        Duration of the run in seconds.
    iterations : int, optional
        Number of iterations.
    custom : Union[Any, Dict[str, Any]], optional
        Custom statistics created by the user. Can normally expect a `Dict[str,
        Any]`.
    """

    duration: Optional[float] = None
    """Duration of the run in seconds."""
    iterations: Optional[int] = None
    """Number of iterations."""
    custom: Optional[
        Union[
            Any,
            Dict[str, Any],
        ]
    ] = None
    """Custom statistics created by the user. Can normally expect a `Dict[str,
    Any]`."""


class ResultStatistics(BaseModel):
    """
    Statistics about a specific result.

    Parameters
    ----------
    duration : float, optional
        Duration of the run in seconds.
    value : float, optional
        Value of the result.
    custom : Union[Any, Dict[str, Any]], optional
        Custom statistics created by the user. Can normally expect a `Dict[str,
        Any]`.
    """

    duration: Optional[float] = None
    """Duration of the run in seconds."""
    value: Optional[float] = None
    """Value of the result."""
    custom: Optional[
        Union[
            Any,
            Dict[str, Any],
        ]
    ] = None
    """Custom statistics created by the user. Can normally expect a `Dict[str,
    Any]`."""


class DataPoint(BaseModel):
    """
    A data point.

    Parameters
    ----------
    x : float
        X coordinate of the data point.
    y : float
        Y coordinate of the data point.
    """

    x: float
    """X coordinate of the data point."""
    y: float
    """Y coordinate of the data point."""


class Series(BaseModel):
    """
    A series of data points.

    Parameters
    ----------
    name : str, optional
        Name of the series.
    data_points : List[DataPoint], optional
        Data of the series.
    """

    name: Optional[str] = None
    """Name of the series."""
    data_points: Optional[List[DataPoint]] = None
    """Data of the series."""


class SeriesData(BaseModel):
    """
    Data of a series.

    Parameters
    ----------
    value : Series, optional
        A series for the value of the solution.
    custom : List[Series], optional
        A list of series for custom statistics.
    """

    value: Optional[Series] = None
    """A series for the value of the solution."""
    custom: Optional[List[Series]] = None
    """A list of series for custom statistics."""


class Statistics(BaseModel):
    """
    Statistics of a solution.

    Parameters
    ----------
    run : RunStatistics, optional
        Statistics about the run.
    result : ResultStatistics, optional
        Statistics about the last result.
    series_data : SeriesData, optional
        Series data about some metric.
    statistics_schema : str, optional
        Schema (version). This class only supports `v1`.
    """

    run: Optional[RunStatistics] = None
    """Statistics about the run."""
    result: Optional[ResultStatistics] = None
    """Statistics about the last result."""
    series_data: Optional[SeriesData] = None
    """Data of the series."""
    statistics_schema: Optional[str] = Field(alias="schema", default="v1")
    """Schema (version). This class only supports `v1`."""



# © 2019-present nextmv.io inc

"""
Check provides a plugin that allows you to check models and solutions.

Checking a model or a solution checks the unplanned plan units. It checks each
individual plan unit if it can be added to the solution. If the plan unit can
be added to the solution, the report will include on how many vehicles and
what the impact would be on the objective value. If the plan unit cannot be
added to the solution, the report will include the reason why it cannot be
added to the solution.

The check can be invoked on a nextroute.Model or a nextroute.Solution. If the
check is invoked on a model, an empty solution is created and the check is
executed on this empty solution. An empty solution is a solution with all the
initial stops that are fixed, initial stops that are not fixed are not added
to the solution. The check is executed on the unplanned plan units of the
solution. If the check is invoked on a solution, it is executed on the
unplanned plan units of the solution.
"""

from .schema import Objective as Objective
from .schema import ObjectiveTerm as ObjectiveTerm
from .schema import Output as Output
from .schema import PlanUnit as PlanUnit
from .schema import Solution as Solution
from .schema import Summary as Summary
from .schema import Vehicle as Vehicle



# © 2019-present nextmv.io inc

"""
This module contains definitions for the schema in the Nextroute check.
"""

from typing import Dict, List, Optional

from nextroute.base_model import BaseModel


class ObjectiveTerm(BaseModel):
    """Check of the individual terms of the objective for a move."""

    base: Optional[float] = None
    """Base of the objective term."""
    factor: Optional[float] = None
    """Factor of the objective term."""
    name: Optional[str] = None
    """Name of the objective term."""
    value: Optional[float] = None
    """Value of the objective term, which is equivalent to `self.base *
    self.factor`."""


class Objective(BaseModel):
    """Estimate of an objective of a move."""

    terms: Optional[List[ObjectiveTerm]] = None
    """Check of the individual terms of the objective."""
    value: Optional[float] = None
    """Value of the objective."""
    vehicle: Optional[str] = None
    """ID of the vehicle for which it reports the objective."""


class Solution(BaseModel):
    """Solution that the check has been executed on."""

    objective: Optional[Objective] = None
    """Objective of the start solution."""
    plan_units_planned: Optional[int] = None
    """Number of plan units planned in the start solution."""
    plan_units_unplanned: Optional[int] = None
    """Number of plan units unplanned in the start solution."""
    stops_planned: Optional[int] = None
    """Number of stops planned in the start solution."""
    vehicles_not_used: Optional[int] = None
    """Number of vehicles not used in the start solution."""
    vehicles_used: Optional[int] = None
    """Number of vehicles used in the start solution."""


class Summary(BaseModel):
    """Summary of the check."""

    moves_failed: Optional[int] = None
    """number of moves that failed. A move can fail if the estimate of a
    constraint is incorrect. A constraint is incorrect if `ModelConstraint.
    EstimateIsViolated` returns true and one of the violation checks returns
    false. Violation checks are implementations of one or more of the
    interfaces [SolutionStopViolationCheck], [SolutionVehicleViolationCheck] or
    [SolutionViolationCheck] on the same constraint. Most constraints do not
    need and do not have violation checks as the estimate is perfect. The
    number of moves failed can be more than one per plan unit as we continue to
    try moves on different vehicles until we find a move that is executable or
    all vehicles have been visited."""
    plan_units_best_move_failed: Optional[int] = None
    """Number of plan units for which the best move can not be planned. This
    should not happen if all the constraints are implemented correct."""
    plan_units_best_move_found: Optional[int] = None
    """Number of plan units for which at least one move has been found and the
    move is executable."""
    plan_units_best_move_increases_objective: Optional[int] = None
    """Number of plan units for which the best move is executable but would
    increase the objective value instead of decreasing it."""
    plan_units_checked: Optional[int] = None
    """Number of plan units that have been checked. If this is less than
    `self.plan_units_to_be_checked` the check timed out."""
    plan_units_have_no_move: Optional[int] = None
    """Number of plan units for which no feasible move has been found. This
    implies there is no move that can be executed without violating a
    constraint."""
    plan_units_to_be_checked: Optional[int] = None
    """Number of plan units to be checked."""


class PlanUnit(BaseModel):
    """Check of a plan unit."""

    best_move_failed: Optional[bool] = None
    """True if the plan unit's best move failed to execute."""
    best_move_increases_objective: Optional[bool] = None
    """True if the best move for the plan unit increases the objective."""
    best_move_objective: Optional[Objective] = None
    """Estimate of the objective of the best move if the plan unit has a best
    move."""
    constraints: Optional[Dict[str, int]] = None
    """Constraints that are violated for the plan unit."""
    has_best_move: Optional[bool] = None
    """True if a move is found for the plan unit. A plan unit has no move found
    if the plan unit is over-constrained or the move found is too expensive."""
    stops: Optional[List[str]] = None
    """IDs of the sops in the plan unit."""
    vehicles_have_moves: Optional[int] = None
    """Number of vehicles that have moves for the plan unit. Only calculated if
    the verbosity is very high."""
    vehicles_with_moves: Optional[List[str]] = None
    """IDs of the vehicles that have moves for the plan unit. Only calculated
    if the verbosity is very high."""


class Vehicle(BaseModel):
    """Check of a vehicle."""

    id: str
    """ID of the vehicle."""

    plan_units_have_moves: Optional[int] = None
    """Number of plan units that have moves for the vehicle. Only calculated if
    the depth is medium."""


class Output(BaseModel):
    """Output of a feasibility check."""

    duration_maximum: Optional[float] = None
    """Maximum duration of the check, in seconds."""
    duration_used: Optional[float] = None
    """Duration used by the check, in seconds."""
    error: Optional[str] = None
    """Error raised during the check."""
    plan_units: Optional[List[PlanUnit]] = None
    """Check of the individual plan units."""
    remark: Optional[str] = None
    """Remark of the check. It can be "ok", "timeout" or anything else that
    should explain itself."""
    solution: Optional[Solution] = None
    """Start soltuion of the check."""
    summary: Optional[Summary] = None
    """Summary of the check."""
    vehicles: Optional[List[Vehicle]] = None
    """Check of the vehicles."""
    verbosity: Optional[str] = None
    """Verbosity level of the check."""



# © 2019-present nextmv.io inc

import os
import platform
import subprocess

from setuptools import Distribution, setup

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class MyWheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False

        def get_tag(self):
            python, abi, plat = _bdist_wheel.get_tag(self)
            # python, abi = "py3", "none"
            return python, abi, plat

    class MyDistribution(Distribution):
        def __init__(self, *attrs):
            Distribution.__init__(self, *attrs)
            self.cmdclass["bdist_wheel"] = MyWheel

        def is_pure(self):
            return False

        def has_ext_modules(self):
            return True

except ImportError:

    class MyDistribution(Distribution):
        def is_pure(self):
            return False

        def has_ext_modules(self):
            return True


# Compile Nextroute binary. We cross-compile (if necessary) for the current
# platform. We also set CGO_ENABLED=0 to ensure that the binary is statically
# linked.
goos = platform.system().lower()
goarch = platform.machine().lower()

if goos not in ["linux", "windows", "darwin"]:
    raise Exception(f"unsupported operating system: {goos}")

# Translate the architecture to the Go convention.
if goarch == "x86_64":
    goarch = "amd64"
elif goarch == "aarch64":
    goarch = "arm64"

if goarch not in ["amd64", "arm64"]:
    raise Exception(f"unsupported architecture: {goarch}")

# Compile the binary.
print(f"Compiling Nextroute binary for {goos} {goarch}...")
cwd = os.getcwd()
standalone_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "cmd")
os.chdir(standalone_dir)
call = ["go", "build", "-o", "../src/nextroute/bin/nextroute.exe", "."]

try:
    subprocess.check_call(
        call,
        env={
            **os.environ,
            "GOOS": goos,
            "GOARCH": goarch,
            "CGO_ENABLED": "0",
        },
    )
finally:
    os.chdir(cwd)


# Get version from version file.
__version__ = "v0.0.0"
exec(open("./src/nextroute/__about__.py").read())

# Setup package.
setup(
    distclass=MyDistribution,
    version=__version__,
)



# © 2019-present nextmv.io inc

import json
import os
import unittest

from nextroute.schema import Input, Stop, Vehicle


class TestInput(unittest.TestCase):
    filepath = os.path.join(os.path.dirname(__file__), "input.json")

    def test_from_json(self):
        with open(self.filepath) as f:
            json_data = json.load(f)

        nextroute_input = Input.from_dict(json_data)
        parsed = nextroute_input.to_dict()

        for s, stop in enumerate(parsed["stops"]):
            original_stop = json_data["stops"][s]
            self.assertEqual(
                stop,
                original_stop,
                f"stop: parsed({stop}) and original ({original_stop}) should be equal",
            )

        for v, vehicle in enumerate(parsed["vehicles"]):
            original_vehicle = json_data["vehicles"][v]
            self.assertEqual(
                vehicle,
                original_vehicle,
                f"vehicle: parsed ({vehicle}) and original ({original_vehicle}) should be equal",
            )

        self.assertEqual(
            parsed["defaults"],
            json_data["defaults"],
            f"defaults: parsed ({parsed['defaults']}) and original ({json_data['defaults']}) should be equal",
        )

    def test_from_dict(self):
        with open(self.filepath) as f:
            json_data = json.load(f)

        nextroute_input = Input.from_dict(json_data)
        stops = nextroute_input.stops
        for stop in stops:
            self.assertTrue(
                isinstance(stop, Stop),
                f"Stop ({stop}) should be of type Stop.",
            )

        vehicles = nextroute_input.vehicles
        for vehicle in vehicles:
            self.assertTrue(
                isinstance(vehicle, Vehicle),
                f"Vehicle ({vehicle}) should be of type Vehicle.",
            )
